/*
  */

#include "ocr.h"

#include <X11/Xlib.h>
#include <X11/extensions/XTest.h>
#include <X11/keysym.h>
#include <X11/Xutil.h>

CanonImage *page_images[MAX_PAGES];
int npages;

static XWindowAttributes attr;
static Window id;
static Display *disp;
static struct timeval tv_startup;
static unsigned wwidth, wheight;
static int wxpos, wypos;

static KeyCode keycode(KeySym sym) {
  return XKeysymToKeycode(disp,sym);
}

#if 0
static void check_exitstatus(int st) {
  eassert(WIFEXITED(st));
  eassert(!WEXITSTATUS(st));
}

static void check_pclose(FILE *f, char *cmd) {
  int r;
  eassert(!ferror(f));
  r= fgetc(f);  eassert(r==EOF);  eassert(feof(f));
  r= pclose(f);  eassert(r>=0);  check_exitstatus(r);
  free(cmd);
}

static CanonImage *screenshot_now(void) {
  char *cmd;
  CanonImage *ci;
  int r;
  
  r= asprintf(&cmd, "xwd -silent -id 0x%lx | xwdtopnm", (unsigned long)id);
  eassert(r>=0);
  FILE *f= popen(cmd,"r");  eassert(f);
  ci= file_read_image(f);
  check_pclose(f, cmd);
  return ci;
}
#endif

static void screenshot_startup(void) {
  int r;
  disp= XOpenDisplay(0);  eassert(disp);
  r= gettimeofday(&tv_startup,0);  eassert(!r);
}

#if 0
static CanonImage *single_page(void) {
  int r;
  r= XRaiseWindow(disp, id);  eassert(r);
  r= XSync(disp, False);  eassert(r);
  return screenshot_now();
}
#endif

/*---------- pager ----------*/

typedef XImage Snapshot;

//static size_t snapshot_alloc= 1024;
static double last_input;
static const double min_update_allowance= 0.25;

static double timestamp(void) {
  struct timeval tv;
  int r;
  
  r= gettimeofday(&tv,0);  eassert(!r);
  double t= (tv.tv_sec - tv_startup.tv_sec) +
            (tv.tv_usec - tv_startup.tv_usec) * 1e-6;
  fprintf(stderr,"%f\n",t);
  return t;
}
static void delay(double need_sleep) {
  int r;
  fprintf(stderr,"PAGING     delay %f\n",need_sleep);
  r= usleep(need_sleep * 1e6);  eassert(!r);
}

static void sync_after_input(void) {
  int r;
  r= XSync(disp, False);  eassert(r);
  last_input= timestamp();
}

static void send_key(KeySym sym) {
  XTestFakeKeyEvent(disp, keycode(sym),1, 10);
  XTestFakeKeyEvent(disp, keycode(sym),0, 10);
}

static void send_pgup_many(void) {
  int i;
  for (i=0; i<25; i++)
    send_key(XK_Prior);
  fprintf(stderr,"PAGING   PageUp x %d\n",i);
  sync_after_input();
}
static void send_pgdown(void) {
  send_key(XK_Next);
  fprintf(stderr,"PAGING   PageDown\n");
  sync_after_input();
}

static void free_snapshot(Snapshot **io) {
  if (*io) XDestroyImage(*io);
  *io= 0;
}

static void snapshot(Snapshot **output) {
//  char *cmd;
//  int r;
//  XImage *xim;
  
  free_snapshot(output);

  fprintf(stderr,"PAGING   snapshot\n");

  timestamp();
  *output= XGetImage(disp,id, 0,0, wwidth,wheight, AllPlanes, ZPixmap);
  timestamp();
  
  fprintf(stderr,"PAGING   snapshot done.\n");
}

static int identical(const Snapshot *a, const Snapshot *b) {
  if (!(a->width == b->width &&
	a->height == b->height &&
	a->bytes_per_line == b->bytes_per_line &&
	a->format == b->format))
    return 0;
  return !memcmp(a->data, b->data, a->bytes_per_line * a->height);
}

static void wait_for_stability(Snapshot **output,
			       const Snapshot *previously,
			       void (*with_keypress)(void)) {
  Snapshot *last=0;
  /* waits longer if we're going to return an image identical to previously
   * if previously==0, all images are considered identical to it */

  fprintf(stderr,"PAGING  wait_for_stability"
	  "  last_input=%f previously=%p\n",
	  last_input, previously);

  for (;;) {
    double at_snapshot= timestamp();
    double need_sleep= min_update_allowance - (at_snapshot - last_input);
    if (need_sleep > 0) { delay(need_sleep); continue; }

    snapshot(output);

    if (!with_keypress &&
	!(previously && identical(*output,previously))) {
      fprintf(stderr,"PAGING  wait_for_stability  simple\n");
      break;
    }

    if (last && identical(*output,last)) {
      fprintf(stderr,"PAGING  wait_for_stability  stabilised\n");
      break;
    }
    
    fprintf(stderr,"PAGING  wait_for_stability  retry\n");

    free_snapshot(&last); last=*output; *output=0;

    if (with_keypress)
      with_keypress();

    delay(0.5);
  }

  free_snapshot(&last);
  fprintf(stderr,"PAGING  wait_for_stability done.\n");
}

static void raise_and_get_details(void) {
  int r;
  int evbase,errbase,majver,minver;
  unsigned bd,depth;
  Window dummy;
  
  fprintf(stderr,"PAGING raise_and_get_details\n");

  r= XTestQueryExtension(disp, &evbase,&errbase,&majver,&minver);
  eassert(r==True);

  r= XRaiseWindow(disp, id);  eassert(r);

  r= XGetWindowAttributes(disp, id, &attr);  eassert(r);
  r= XGetGeometry(disp,id, &attr.root,
		  &wxpos,&wypos, &wwidth,&wheight,
		  &bd,&depth);
  eassert(r);

  r= XTranslateCoordinates(disp, id,attr.root, 160,160, &wxpos,&wypos,
			   &dummy);
  eassert(r);
}

static void set_focus(void) {
  int screen= XScreenNumberOfScreen(attr.screen);

  fprintf(stderr,"PAGING set_focus\n");

  XTestFakeMotionEvent(disp,screen, wxpos,wypos, 0);

  XTestFakeButtonEvent(disp,1,1, 50);
  XTestFakeButtonEvent(disp,1,0, 50);

  sync_after_input();
  fprintf(stderr,"PAGING raise_and_set_focus done.\n");
}

typedef struct {
  unsigned long mask;
  int lshift, rshift;
} ShMask;

static void compute_shift_mask(ShMask *sm, int targshift,
			       unsigned long ximage_mask) {
  unsigned long below;
  
  sm->lshift= 0;
  sm->rshift= 0;
  sm->mask= 0xfful << targshift;
  below= ~0ul << targshift;
  
  for (;;) {
    if (ximage_mask < sm->mask) {
      sm->lshift++;  ximage_mask <<= 1;
    } else if ((ximage_mask & ~below) > sm->mask) {
      sm->rshift++;  ximage_mask >>= 1;
    } else {
      break;
    }
    assert(!(sm->lshift && sm->rshift));
  }
  assert(sm->lshift < LONG_BIT);
  assert(sm->rshift < LONG_BIT);
}

static CanonImage *convert_page(Snapshot *sn) {
  ShMask shiftmasks[3];
  CanonImage *im;

#define COMPUTE_SHIFT_MASK(ix, targshift, rgb) \
  compute_shift_mask(&shiftmasks[ix], targshift, sn->rgb##_mask)
  COMPUTE_SHIFT_MASK(0, 16, red);
  COMPUTE_SHIFT_MASK(1, 8,  green);
  COMPUTE_SHIFT_MASK(2, 0,  blue);

  CANONICALISE_IMAGE(im, sn->width, sn->height, {
    long xrgb= XGetPixel(sn, x, y);
    int i;
    rgb= 0;
    for (i=0; i<3; i++)
      rgb |= ((xrgb << shiftmasks[i].lshift)
	      >> shiftmasks[i].rshift) & shiftmasks[i].mask;
  });

  return im;
}

static void read_pages(void) {
  Snapshot *current=0, *last=0;
  CanonImage *test;

  /* find the window and check it's on the right kind of screen */
  raise_and_get_details();
  snapshot(&current);
  test= convert_page(current);
  find_structure(test);
  free(test);

  /* page to the top - keep pressing page up until the image stops changing */
  set_focus();
  wait_for_stability(&current,0, send_pgup_many);

  /* now to actually page down */
  for (;;) {
    fprintf(stderr,"paging page %d\n",npages);

    eassert(npages < MAX_PAGES);
    page_images[npages]= convert_page(current);
    free_snapshot(&last); last=current; current=0;

    fprintf(stderr,"PAGING page %d converted\n",npages);

    wait_for_stability(&current,last, 0);
    if (npages &&  /* first pagedown doesn't do much */
	identical(current,last)) {
      free_snapshot(&current);
      break;
    }

    send_pgdown();
    npages++;
  }
  fprintf(stderr,"PAGING all done.\n");
}    

int main(int argc, char **argv) {
  screenshot_startup();

  id= strtoul(*++argv,0,0);

  read_pages();
  return 0;
}

